import logging
import queue
import threading
from collections.abc import Iterator
from typing import Any, Callable, Optional

import h5py
import numpy as np

from pandablocks.commands import Arm

from .asyncio import AsyncioClient
from .connections import GATE_DURATION_FIELD, SAMPLES_FIELD
from .responses import EndData, EndReason, FieldCapture, FrameData, ReadyData, StartData

# Define the public API of this module
__all__ = [
    "Pipeline",
    "HDFWriter",
    "FrameProcessor",
    "HDFDataOverrunException",
    "create_pipeline",
    "create_default_pipeline",
    "stop_pipeline",
    "write_hdf_files",
]


class HDFDataOverrunException(Exception):
    """Raised if `DATA_OVERRUN` occurs while receiving data for HDF file"""


class Stop:
    def __str__(self) -> "str":
        return "<Stop>"

    __repr__ = __str__


STOP = Stop()


class Pipeline(threading.Thread):
    """Helper class that runs a pipeline consumer process in its own thread"""

    #: Subclasses should create this dictionary with handlers for each data
    #: type, returning transformed data that should be passed downstream
    what_to_do: dict[type, Callable]
    downstream: Optional["Pipeline"] = None

    def __init__(self):
        super().__init__()
        self.queue: queue.Queue[Any] = queue.Queue()  # type: ignore

    def run(self):
        while True:
            data = self.queue.get()
            if data is STOP:
                # stop() called below
                break
            else:
                func = self.what_to_do.get(type(data), None)
                if func:
                    # If we have a handler, use it to transform the data
                    data = func(data)
                if self.downstream and data is not None:
                    # Pass the (possibly transformed) data downstream
                    self.downstream.queue.put_nowait(data)

    def stop(self):
        """Stop the processing after the current queue has been emptied"""
        self.queue.put(STOP)


class HDFWriter(Pipeline):
    """Write an HDF file per data collection. Each field will be
    written in a 1D dataset ``/<field.name>.<field.capture>``.

    Args:
        file_names: Iterator of file names. Must be full file paths. Will be called once
            per file created.
        capture_record_hdf_names: A dictionary of alternate dataset names to use for
            each field. For example

            .. code-block:: python

                {
                    "COUNTER1.OUT": {
                        "Value": "name",
                        "Min": "name-min",
                        "Max": "name-max"
                    }
                }
    """

    def __init__(
        self,
        file_names: Iterator[str],
        capture_record_hdf_names: dict[str, dict[str, str]],
    ):
        super().__init__()
        self.file_names = file_names
        self.hdf_file: Optional[h5py.File] = None
        self.datasets: list[h5py.Dataset] = []
        self.capture_record_hdf_names = capture_record_hdf_names
        self.what_to_do = {
            StartData: self.open_file,
            list: self.write_frame,
            EndData: self.close_file,
        }

    def create_dataset(self, field: FieldCapture, raw: bool):
        # Data written in a big stack, growing in that dimension
        assert self.hdf_file, "File not open yet"

        dataset_name = self.capture_record_hdf_names.get(field.name, {}).get(
            field.capture, f"{field.name}.{field.capture}"
        )

        dtype = field.raw_mode_dataset_dtype if raw else field.type

        return self.hdf_file.create_dataset(
            f"/{dataset_name}",
            dtype=dtype,
            shape=(0,),
            maxshape=(None,),
        )

    def open_file(self, data: StartData):
        try:
            self.file_path = next(self.file_names)
        except IndexError:
            logging.exception(
                "Not enough file names available when opening new HDF5 file"
            )
            raise
        self.hdf_file = h5py.File(self.file_path, "w", libver="latest")
        raw = data.process == "Raw"
        self.datasets = [self.create_dataset(field, raw) for field in data.fields]
        self.hdf_file.swmr_mode = True

        # Save parameters
        if data.arm_time is not None:
            self.hdf_file.attrs["arm_time"] = data.arm_time
        if data.start_time is not None:
            self.hdf_file.attrs["start_time"] = data.start_time
        if data.hw_time_offset_ns is not None:
            self.hdf_file.attrs["hw_time_offset_ns"] = data.hw_time_offset_ns

        logging.info(
            f"Opened '{self.file_path}' with {data.sample_bytes} byte samples "
            f"stored in {len(self.datasets)} datasets"
        )

    def write_frame(self, data: list[np.ndarray]):
        for dataset, column in zip(self.datasets, data):
            # Append to the end, flush when done
            written = dataset.shape[0]
            dataset.resize((written + column.shape[0],))
            dataset[written:] = column
            dataset.flush()

        # Return the number of samples written
        return dataset.shape[0]

    def close_file(self, data: EndData):
        assert self.hdf_file, "File not open yet"
        self.hdf_file.close()
        self.hdf_file = None
        logging.info(
            f"Closed '{self.file_path}' after receiving {data.samples} "
            f"samples. End reason is '{data.reason.value}'"
        )
        self.file_path = ""


class FrameProcessor(Pipeline):
    """Scale field data according to the information in the StartData"""

    def __init__(self) -> None:
        super().__init__()
        self.processors: list[Callable] = []
        self.what_to_do = {
            StartData: self.create_processors,
            FrameData: self.scale_data,
        }

    def create_processor(self, field: FieldCapture, raw: bool):
        column_name = f"{field.name}.{field.capture}"

        if raw and field.capture == "Mean":

            def mean_callable(data):
                if GATE_DURATION_FIELD in data.dtype.names:
                    gate_duration = data[GATE_DURATION_FIELD]
                else:
                    gate_duration = data[SAMPLES_FIELD]

                return (data[column_name] * field.scale / gate_duration) + field.offset

            return mean_callable
        elif raw and field.has_scale_or_offset:
            return lambda data: data[column_name] * field.scale + field.offset
        else:
            return lambda data: data[column_name]

    def create_processors(self, data: StartData) -> StartData:
        raw = data.process == "Raw"
        self.processors = [self.create_processor(field, raw) for field in data.fields]
        return data

    def scale_data(self, data: FrameData) -> list[np.ndarray]:
        return [process(data.data) for process in self.processors]


def create_default_pipeline(
    file_names: Iterator[str],
    capture_record_hdf_names: dict[str, dict[str, str]],
    *additional_downstream_pipelines: Pipeline,
) -> list[Pipeline]:
    """Create the default processing pipeline consisting of one `FrameProcessor` and
    one `HDFWriter`. See `create_pipeline` for more details.

    Args:
        file_names: Iterator of file names. Must be full file paths. Will be called once
            per file created. As required by `HDFWriter`.
        capture_record_hdf_names: A dictionary of dataset names to use for each field.
            The keys are record names, the values are another dictionary of
            capture type to dataset name.
        additional_downstream_pipelines: Any number of additional pipelines to add
            downstream.
    """

    return create_pipeline(
        FrameProcessor(),
        HDFWriter(file_names, capture_record_hdf_names),
        *additional_downstream_pipelines,
    )


def create_pipeline(*elements: Pipeline) -> list[Pipeline]:
    """Create a pipeline of elements, wiring them and starting them before
    returning them"""
    pipeline: list[Pipeline] = []
    for element in elements:
        if pipeline:
            pipeline[-1].downstream = element
        pipeline.append(element)
        element.start()
    return pipeline


def stop_pipeline(pipeline: list[Pipeline]):
    """Stop and join each element of the pipeline"""
    for element in pipeline:
        # Note that we stop and join each element in turn.
        # This ensures all data is flushed all the way down
        # even if there is lots left in a queue
        element.stop()
        element.join()


async def write_hdf_files(
    client: AsyncioClient,
    file_names: Iterator[str],
    num: int = 1,
    arm: bool = False,
    flush_period: float = 1,
):
    """Connect to host PandA data port, and write num acquisitions
    to HDF file according to scheme

    Args:
        client: The `AsyncioClient` to use for communications
        file_names: Iterator of file names. Must be full file paths. Will be called once
            per file created.
        num: The number of acquisitions to store in separate files. 0 = Infinite capture
        arm: Whether to arm PCAP at the start, and after each successful acquisition

    Raises:
        HDFDataOverrunException: if there is a data overrun.
    """
    counter = 0

    end_data = None
    pipeline = create_default_pipeline(file_names, {})
    try:
        async for data in client.data(scaled=False, flush_period=flush_period):
            pipeline[0].queue.put_nowait(data)
            if isinstance(data, EndData):
                end_data = data
                counter += 1
                if counter == num:
                    # We produced the right number of frames
                    break
            if type(data) in (ReadyData, EndData) and arm:
                # Told to arm at the beginning, and after each acquisition ends
                await client.send(Arm())
        if end_data and end_data.reason == EndReason.DATA_OVERRUN:
            raise HDFDataOverrunException(
                "Data overrun - streaming aborted! Last frame may be corrupt."
            )
    finally:
        stop_pipeline(pipeline)



import sys

import h5py
import matplotlib.pyplot as plt

if __name__ == "__main__":
    with h5py.File(sys.argv[1], "r") as hdf:
        # Print the dataset names in the file
        datasets = list(hdf)
        print(datasets)
        # Plot the counter mean values for the 3 counters we know are captured
        for i in range(1, 4):
            plt.plot(hdf[f"COUNTER{i}.OUT.Mean"], label=f"Counter {i}")
        # Add a legend and show the plot
        plt.legend()
        plt.show()



from collections.abc import Generator
from typing import TypeVar, Union

T = TypeVar("T")


class Exchange:
    """A helper class representing the lines to send to PandA and
    the lines received"""

    def __init__(self, to_send: Union[str, list[str]]):
        if isinstance(to_send, str):
            self.to_send = [to_send]
        else:
            self.to_send = to_send
        self.received: list[str] = []
        self.is_multiline = False

    def _error_message(self) -> str:
        sent = "\n".join(self.to_send)
        received = "\n".join(self.received)
        return f"{sent!r} -> {received!r}"

    def check_ok(self):
        assert self.received == ["OK"], self._error_message()

    @property
    def line(self) -> str:
        """Check received is not multiline and return the line"""
        assert not self.is_multiline and self.received[0].startswith(
            "OK ="
        ), self._error_message()
        # Remove the OK= header
        return self.received[0][4:]

    @property
    def multiline(self) -> list[str]:
        """Return the multiline received lines, processed to remove markup"""
        assert self.is_multiline, self._error_message()
        # Remove the ! and . markup
        return [line[1:] for line in self.received[:-1]]


Exchanges = Union[Exchange, list[Exchange]]
ExchangeGenerator = Generator[Exchanges, None, T]



import queue
from pathlib import Path

import numpy as np
import pytest

from pandablocks.hdf import HDFWriter, Pipeline, create_default_pipeline, stop_pipeline
from pandablocks.responses import EndData, EndReason, FieldCapture, FrameData, StartData


def test_pipeline_returns_number_written(tmp_path):
    NUMBER_OF_FRAMES_WRITTEN = 10000

    num_written_queue = queue.Queue()

    class DummyDownstream(Pipeline):
        def __init__(self):
            self.what_to_do = {int: num_written_queue.put_nowait}
            super().__init__()

    file_counter = DummyDownstream()

    try:
        pipeline = create_default_pipeline(
            iter([Path(tmp_path / "1.h5")]), {}, file_counter
        )

        pipeline[0].queue.put_nowait(
            StartData(
                [
                    FieldCapture(
                        name="COUNTER1.OUT",
                        type=np.dtype("float64"),
                        capture="Value",
                        scale=1,
                        offset=0,
                        units="",
                    )
                ],
                0,
                "Scaled",
                "Framed",
                52,
                "2024-03-05T20:27:12.607841574Z",
                "2024-03-05T20:27:12.608875498Z",
                100555,
            ),
        )
        pipeline[0].queue.put_nowait(
            FrameData(
                np.array(
                    NUMBER_OF_FRAMES_WRITTEN * [(1,)],
                    dtype=[("COUNTER1.OUT.Value", "<f8")],
                )
            )
        )
        pipeline[0].queue.put_nowait(EndData(5, EndReason.DISARMED))

        assert num_written_queue.get() == NUMBER_OF_FRAMES_WRITTEN
    finally:
        stop_pipeline(pipeline)


def test_field_capture_pcap_bits():
    pcap_bits_frame_data = FieldCapture(
        name="PCAP.BITS",
        type=np.dtype("uint32"),
        capture="Value",
        scale=None,
        offset=None,
        units=None,
    )

    assert not pcap_bits_frame_data.has_scale_or_offset
    assert pcap_bits_frame_data.raw_mode_dataset_dtype is np.dtype("uint32")

    frame_data_without_scale_offset = FieldCapture(
        name="frame_data_without_scale_offset",
        type=np.dtype("uint32"),
        capture="Value",
        scale=1.0,
        offset=0.0,
        units="",
    )

    assert not frame_data_without_scale_offset.has_scale_or_offset
    assert frame_data_without_scale_offset.raw_mode_dataset_dtype is np.dtype("float64")

    with pytest.raises(
        ValueError,
        match=(
            "If any of `scale=None`, `offset=0.0`, or "
            "`units=` is set, all must be set"
        ),
    ):
        _ = FieldCapture(
            name="malformed_frame_data",
            type=np.dtype("uint32"),
            capture="Value",
            scale=None,
            offset=0.0,
            units="",
        )

    frame_data_with_offset = FieldCapture(
        name="frame_data_with_offset",
        type=np.dtype("float64"),
        capture="Value",
        scale=1.0,
        offset=1.0,
        units="",
    )
    frame_data_with_scale = FieldCapture(
        name="frame_data_with_scale",
        type=np.dtype("float64"),
        capture="Value",
        scale=1.1,
        offset=0.0,
        units="",
    )

    assert frame_data_with_offset.has_scale_or_offset
    assert frame_data_with_offset.raw_mode_dataset_dtype is np.dtype("float64")
    assert frame_data_with_scale.has_scale_or_offset
    assert frame_data_with_scale.raw_mode_dataset_dtype is np.dtype("float64")

    frame_data_with_scale_and_offset = FieldCapture(
        name="frame_data_with_scale_and_offset",
        type=np.dtype("float64"),
        capture="Value",
        scale=1.1,
        offset=0.0,
        units="",
    )

    assert frame_data_with_scale_and_offset.has_scale_or_offset
    assert frame_data_with_scale_and_offset.raw_mode_dataset_dtype is np.dtype(
        "float64"
    )


@pytest.mark.parametrize(
    "capture_record_hdf_names,expected_names",
    [
        (
            {},
            {
                "/COUNTER1.OUT.Value",
                "/COUNTER2.OUT.Mean",
                "/COUNTER2.OUT.Max",
                "/COUNTER2.OUT.Min",
            },
        ),
        (
            {"COUNTER1.OUT": {"Value": "scientific-name"}},
            {
                "/scientific-name",
                "/COUNTER2.OUT.Mean",
                "/COUNTER2.OUT.Max",
                "/COUNTER2.OUT.Min",
            },
        ),
        (
            {
                "COUNTER2.OUT": {
                    "Mean": "scientific-name",
                    "Max": "scientific-name-max",
                    "Min": "scientific-name-min",
                }
            },
            {
                "/COUNTER1.OUT.Value",
                "/scientific-name",
                "/scientific-name-max",
                "/scientific-name-min",
            },
        ),
        (
            {
                "COUNTER1.OUT": {"Value": "scientific-name1"},
                "COUNTER2.OUT": {
                    "Mean": "scientific-name2",
                    "Max": "scientific-name2-max",
                    "Min": "scientific-name2-min",
                },
            },
            {
                "/scientific-name1",
                "/scientific-name2",
                "/scientific-name2-max",
                "/scientific-name2-min",
            },
        ),
    ],
)
def test_hdf_writer_uses_alternative_dataset_names(
    tmp_path, capture_record_hdf_names, expected_names
):
    hdf_writer = HDFWriter(
        iter([str(tmp_path / "test_file.h5")]), capture_record_hdf_names
    )

    start_data = StartData(
        fields=[
            FieldCapture(
                name="COUNTER1.OUT",
                type=np.dtype("float64"),
                capture="Value",
                scale=1,
                offset=0,
                units="",
            ),
            FieldCapture(
                name="COUNTER2.OUT",
                type=np.dtype("float64"),
                capture="Min",
                scale=1,
                offset=0,
                units="",
            ),
            FieldCapture(
                name="COUNTER2.OUT",
                type=np.dtype("float64"),
                capture="Max",
                scale=1,
                offset=0,
                units="",
            ),
            FieldCapture(
                name="COUNTER2.OUT",
                type=np.dtype("float64"),
                capture="Mean",
                scale=1,
                offset=0,
                units="",
            ),
        ],
        missed=0,
        process="Scaled",
        format="Framed",
        sample_bytes=52,
        arm_time="2024-03-05T20:27:12.607841574Z",
        start_time="2024-03-05T20:27:12.608875498Z",
        hw_time_offset_ns=100555,
    )

    hdf_writer.open_file(start_data)

    assert {dataset.name for dataset in hdf_writer.datasets} == expected_names



import pytest_asyncio

from pandablocks._control import BlockCompleter
from pandablocks.blocking import BlockingClient


@pytest_asyncio.fixture
def dummy_server_with_blocks(dummy_server_in_thread):
    dummy_server_in_thread.send += [
        "!PCAP 1\n!LUT 8\n!SRGATE 2\n.",
        "!INPB 1 bit_mux\n!TYPEA 5 param enum\n.",  # LUT fields
        "!TRIG_EDGE 3 param enum\n!GATE 1 bit_mux\n.",  # PCAP fields
        "!OUT 1 bit_out\n.",  # SRGATE fields
    ]
    yield dummy_server_in_thread


def test_complete_one_block(dummy_server_with_blocks):
    with BlockingClient("localhost") as client:
        completer = BlockCompleter(client)
        assert completer("PCA", 0) == "PCAP"
        assert completer.matches == ["PCAP"]
        assert completer("PCAP.", 0) == "PCAP.GATE"
        assert completer.matches == ["PCAP.GATE", "PCAP.TRIG_EDGE"]
        completer("LU", 0)
        assert completer.matches == [f"LUT{i}" for i in range(1, 9)]
        completer("LUT5.", 0)
        assert completer.matches == ["LUT5.INPB", "LUT5.TYPEA"]
        completer("LUT5.TYPEA_IN", 0)
        assert completer.matches == []
        completer("LUT5.TYPE", 0)
        assert completer.matches == ["LUT5.TYPEA"]
        assert completer("LUT5.TYPEA_INP?", 0) is None


def test_complete_stars(dummy_server_with_blocks):
    with BlockingClient("localhost") as client:
        completer = BlockCompleter(client)
        assert completer("*PCA", 0) == "*PCAP.STATUS?"
        assert completer.matches == [
            "*PCAP.STATUS?",
            "*PCAP.CAPTURED?",
            "*PCAP.COMPLETION?",
            "*PCAP.ARM=",
            "*PCAP.DISARM=",
        ]
        assert completer("*DE", 0) == "*DESC.LUT"
        assert completer.matches == ["*DESC.LUT", "*DESC.PCAP", "*DESC.SRGATE"]
        assert completer("*DESC.LUT.", 0) == "*DESC.LUT.INPB"
        assert completer.matches == ["*DESC.LUT.INPB", "*DESC.LUT.TYPEA"]
        assert completer("*ENUMS.LU", 0) == "*ENUMS.LUT"
        assert completer.matches == ["*ENUMS.LUT"]



from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

import numpy as np

# Define the public API of this module
__all__ = [
    "BlockInfo",
    "FieldInfo",
    "UintFieldInfo",
    "ScalarFieldInfo",
    "TimeFieldInfo",
    "SubtypeTimeFieldInfo",
    "EnumFieldInfo",
    "BitOutFieldInfo",
    "BitMuxFieldInfo",
    "PosMuxFieldInfo",
    "TableFieldDetails",
    "TableFieldInfo",
    "PosOutFieldInfo",
    "ExtOutFieldInfo",
    "ExtOutBitsFieldInfo",
    "Changes",
    "EndReason",
    "FieldCapture",
    "Data",
    "ReadyData",
    "StartData",
    "FrameData",
    "EndData",
]

# Control


@dataclass
class BlockInfo:
    """Block number and description as exposed by the TCP server

    Attributes:
        number: The index of this block
        description: The description for this block"""

    number: int = 0
    description: Optional[str] = None


@dataclass
class FieldInfo:
    """Field type, subtype, description and labels as exposed by TCP server:
    https://pandablocks-server.readthedocs.io/en/latest/fields.html#field-types

    Note that many fields will use a more specialised subclass of FieldInfo for
    their additional attributes.

    Attributes:
        type: Field type, like "param", "bit_out", "pos_mux", etc.
        subtype: Some types have subtype, like "uint", "scalar", "lut", etc.
        description: A description of the field
        labels: A list of the valid values for the field when there is a defined list
            of valid values, e.g. those with sub-type "enum"
    """

    type: str
    subtype: Optional[str]
    description: Optional[str]


@dataclass
class UintFieldInfo(FieldInfo):
    """Extended `FieldInfo` for fields with type "param","read", or "write" and subtype
    "uint"""

    max_val: int


@dataclass
class ScalarFieldInfo(FieldInfo):
    """Extended `FieldInfo` for fields with type "param","read", or "write" and subtype
    "scalar"""

    units: str
    scale: float
    offset: float


@dataclass
class TimeFieldInfo(FieldInfo):
    """Extended `FieldInfo` for fields with type "time"""

    units_labels: list[str]


@dataclass
class SubtypeTimeFieldInfo(FieldInfo):
    """Extended `FieldInfo` for fields with type "param","read", or "write" and subtype
    "time"""

    units_labels: list[str]


@dataclass
class EnumFieldInfo(FieldInfo):
    """Extended `FieldInfo` for fields with type "param","read", or "write" and subtype
    "enum"""

    labels: list[str]


@dataclass
class BitOutFieldInfo(FieldInfo):
    """Extended `FieldInfo` for fields with type "bit_out"""

    capture_word: str
    offset: int


@dataclass
class BitMuxFieldInfo(FieldInfo):
    """Extended `FieldInfo` for fields with type "bit_mux"""

    max_delay: int
    labels: list[str]


@dataclass
class PosMuxFieldInfo(FieldInfo):
    """Extended `FieldInfo` for fields with type "pos_mux"""

    labels: list[str]


@dataclass
class TableFieldDetails:
    """Info for each field in a table"""

    subtype: str
    bit_low: int
    bit_high: int
    description: Optional[str] = None
    labels: Optional[list[str]] = None


@dataclass
class TableFieldInfo(FieldInfo):
    """Extended `FieldInfo` for fields with type "table"""

    max_length: int
    fields: dict[str, TableFieldDetails]
    row_words: int


@dataclass
class PosOutFieldInfo(FieldInfo):
    """Extended `FieldInfo` for fields with type "pos_out"""

    capture_labels: list[str]


@dataclass
class ExtOutFieldInfo(FieldInfo):
    """Extended `FieldInfo` for fields with type "ext_out" and subtypes "timestamp"
    or "samples"""

    capture_labels: list[str]


@dataclass
class ExtOutBitsFieldInfo(ExtOutFieldInfo):
    """Extended `ExtOutFieldInfo` for fields with type "ext_out" and subtype "bits"""

    bits: list[str]


@dataclass
class Changes:
    """The changes returned from a ``*CHANGES`` command"""

    #: Map field -> value for single-line values that were returned
    values: dict[str, str]
    #: The fields that were present but without value
    no_value: list[str]
    #: The fields that were in error
    in_error: list[str]
    #: Map field -> value for multi-line values that were returned
    multiline_values: dict[str, list[str]]


# Data


class EndReason(Enum):
    """The reason that a PCAP acquisition completed"""

    #: Experiment completed by falling edge of ``PCAP.ENABLE```
    OK = "Ok"
    #: Client disconnect detected
    EARLY_DISCONNECT = "Early disconnect"
    #: Client not taking data quickly or network congestion, internal buffer overflow.
    #: In raw unscaled mode (i.e., no server-side scaling), the most recent
    #: `FrameData` is likely corrupted.
    DATA_OVERRUN = "Data overrun"
    #: Triggers too fast for configured data capture
    FRAMING_ERROR = "Framing error"
    #: Probable CPU overload on PandA, should not occur
    DRIVER_DATA_OVERRUN = "Driver data overrun"
    #: Data capture too fast for memory bandwidth
    DMA_DATA_ERROR = "DMA data error"
    # Reasons below this point are not from the server, they are generated in code
    #: An unknown exception occurred during HDF5 file processing
    UNKNOWN_EXCEPTION = "Unknown exception"
    #: StartData packets did not match when trying to continue printing to a file
    START_DATA_MISMATCH = "Start Data mismatched"
    #: Experiment manually completed by ``DATA:CAPTURE``
    MANUALLY_STOPPED = "Manually stopped"
    #: Experiment manually completed by ``*PCAP.DISARM=`` command
    DISARMED = "Disarmed"


@dataclass
class FieldCapture:
    """Information about a field that is being captured

    If scale, offset, and units are all `None`, then the field is a
    ``PCAP.BITS``.

    Attributes:
        name: Name of captured field
        type: Numpy data type of the field as transmitted
        capture: Value of CAPTURE field used to enable this field
        scale: Scaling factor
        offset: Offset
        units: Units string
    """

    name: str
    type: np.dtype
    capture: str
    scale: Optional[float] = field(default=None)
    offset: Optional[float] = field(default=None)
    units: Optional[str] = field(default=None)

    def __post_init__(self):
        sou = (self.scale, self.offset, self.units)
        if sou != (None, None, None) and None in sou:
            raise ValueError(
                f"If any of `scale={self.scale}`, `offset={self.offset}`"
                f", or `units={self.units}` is set, all must be set."
            )

    @property
    def raw_mode_dataset_dtype(self) -> np.dtype:
        """We use double for all dtypes that have scale and offset."""
        if self.scale is not None and self.offset is not None:
            return np.dtype("float64")
        return self.type

    @property
    def has_scale_or_offset(self) -> bool:
        """Return True if this field is a PCAP.BITS or PCAP.SAMPLES field"""
        return (self.scale is not None and self.offset is not None) and (
            self.scale != 1 or self.offset != 0
        )


class Data:
    """Baseclass for all responses yielded by a `DataConnection`"""


@dataclass
class ReadyData(Data):
    """Yielded once when the connection is established and ready to take data"""


@dataclass
class StartData(Data):
    """Yielded when a new PCAP acquisition starts.

    Attributes:
        fields: Information about each captured field as a `FieldCapture` object
        missed: Number of samples missed by late data port connection
        process: Data processing option, only "Scaled" or "Raw" are requested
        format: Data delivery formatting, only "Framed" is requested
        sample_bytes: Number of bytes in one sample
    """

    fields: list[FieldCapture]
    missed: int
    process: str
    format: str
    sample_bytes: int
    arm_time: Optional[str]
    start_time: Optional[str]
    hw_time_offset_ns: Optional[int]


@dataclass
class FrameData(Data):
    """Yielded when a new data frame is flushed.

    Attributes:
        data: A numpy `Structured Array <structured_arrays>`

    Data is structured into complete columns. Each column name is
    ``<name>.<capture>`` from the corresponding `FieldInfo`. Data
    can be accessed with these column names. For example::

        # Table view with 2 captured fields
        >>> import numpy
        >>> data = numpy.array([(0, 10),
        ...       (1, 11),
        ...       (2, 12)],
        ...      dtype=[('COUNTER1.OUT.Value', '<f8'), ('COUNTER2.OUT.Value', '<f8')])
        >>> fdata = FrameData(data)
        >>> (fdata.data[0]['COUNTER1.OUT.Value'], fdata.data[0]['COUNTER2.OUT.Value'])
        (np.float64(0.0), np.float64(10.0))
        >>> fdata.column_names # Column names
        ('COUNTER1.OUT.Value', 'COUNTER2.OUT.Value')
        >>> fdata.data['COUNTER1.OUT.Value'] # Column view
        array([0., 1., 2.])
    """

    data: np.ndarray

    @property
    def column_names(self) -> tuple[str, ...]:
        """Return all the column names"""
        names = self.data.dtype.names
        assert names, f"No column names for {self.data.dtype}"
        return names


@dataclass
class EndData(Data):
    """Yielded when a PCAP acquisition ends.

    Attributes:
        samples: The total number of samples (rows) that were yielded
        reason: The `EndReason` for the end of acquisition
    """

    samples: int
    reason: EndReason



import socket
from collections.abc import Iterable, Iterator
from typing import Optional, Union, overload

from .commands import Command, T
from .connections import ControlConnection, DataConnection
from .responses import Data

# Define the public API of this module
__all__ = ["BlockingClient"]


class _SocketHelper:
    _socket: Optional[socket.socket] = None

    def connect(self, host: str, port: int):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((host, port))
        s.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
        self._socket = s

    @property
    def socket(self) -> socket.socket:
        assert self._socket, "connect() not called yet"
        return self._socket

    def close(self):
        s = self.socket
        self._socket = None
        s.shutdown(socket.SHUT_WR)
        s.close()


class BlockingClient:
    """Blocking implementation of a PandABlocks client.
    For example::

        with BlockingClient("hostname-or-ip") as client:
            # Control port is now connected
            resp1, resp2 = client.send([cmd1, cmd2])
            resp3 = client.send(cmd3)
            for data in client.data():
                handle(data)
        # Control and data ports are now disconnected
    """

    def __init__(self, host: str):
        self._host = host
        self._ctrl_connection = ControlConnection()
        self._ctrl_socket = _SocketHelper()

    def connect(self):
        """Connect to the control port, and be ready to handle commands"""
        self._ctrl_socket.connect(self._host, 8888)

    def close(self):
        """Close the control connection, and wait for completion"""
        self._ctrl_socket.close()

    def __enter__(self) -> "BlockingClient":
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @overload
    def send(self, commands: Command[T], timeout: Optional[int] = None) -> T: ...

    @overload
    def send(
        self, commands: Iterable[Command], timeout: Optional[int] = None
    ) -> list: ...

    def send(
        self,
        commands: Union[Command[T], Iterable[Command]],
        timeout: Optional[int] = None,
    ):
        """Send a command to control port of the PandA, returning its response.

        Args:
            commands: If single `Command`, return its response. If a list of commands
                return a list of reponses
            timeout: If no reponse in this time, raise `socket.timeout`
        """
        s = self._ctrl_socket.socket
        s.settimeout(timeout)
        if isinstance(commands, Command):
            commands = [commands]
        else:
            commands = list(commands)
        for command in commands:
            to_send = self._ctrl_connection.send(command)
            s.sendall(to_send)
        # Rely on dicts being ordered, Ellipsis is shorthand for "no response yet"
        cr = {id(command): ... for command in commands}
        while ... in cr.values():
            received = s.recv(4096)
            to_send = self._ctrl_connection.receive_bytes(received)
            s.sendall(to_send)
            for command, response in self._ctrl_connection.responses():
                assert cr[id(command)] is ..., "Already got response for {command}"
                cr[id(command)] = response
        responses = list(cr.values())
        for response in responses:
            if isinstance(response, Exception):
                raise response
        if len(responses) == 1:
            return responses[0]
        else:
            return responses

    def data(
        self, scaled: bool = True, frame_timeout: Optional[int] = None
    ) -> Iterator[Data]:
        """Connect to data port and yield data frames

        Args:
            scaled: Whether to scale and average data frames, reduces throughput
            frame_timeout: If no data is received for this amount of time, raise
                `socket.timeout`
        """
        data_socket = _SocketHelper()
        data_socket.connect(self._host, 8889)

        connection = DataConnection()
        s = data_socket.socket
        s.settimeout(frame_timeout)  # close enough
        s.sendall(connection.connect(scaled))
        try:
            while True:
                received = s.recv(4096)
                yield from connection.receive_bytes(received)
        finally:
            data_socket.close()
